/*
   AGP Memory (ARB_vertex_buffer_object) Allocating Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to use AGP memory when rendering our 3D OpenGL
   scenes.  Why use AGP memory directly if you don't have to you ask?  Well using
   AGP directly will have an extreme increase in performance than without!  To
   use AGP memory we have a choice.  First there is the nvidia way using their extensions.
   The second way is the ATI way which differs of course.  The thrid way and the way most
   people will probably go is the ARB way which is supported by all cards that support AGP
   memory and OpenGL.  I'll try to cover each in their own demo.

   In this demo we are using the ARB_vertex_buffer_object extension.  This demo was
   meant to be viewed after the first AGP memory demo.

   Here are the function prototypes for your convience for all AGP memory functions we use...
   (the last one I didn't find the prototype for so I filled it in based on how I used it)

   glBindBufferARB(GLenum target, GLuint buffer);
   glDeleteBuffersARB(GLsizei n, const GLuint *buffers);
   glGenBuffersARB(GLsizei n, GLuint *buffers);
   glBufferDataARB(GLenum target, GLsizeiptrARB size, const GLvoid *data, GLenum usage);
   glGetBufferParameterivARB(GLenum target, GLenum target, GLuint *size);
*/


// Demo: AGP Memory (ARB_vertex_buffer_object) Allocating.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


#include<windows.h>        // Windows header file.
#include<stdlib.h>         // Standard library.
#include<gl/gl.h>          // OpenGL header.
#include<gl/glu.h>         // OpenGL utility.
#include<gl/glext.h>       // OpenGL extensions.
#include"CTGALoader.h"     // Holds the tga loading class.


// Function prototypes...
bool InitializeGL();       // Initialize OpenGL.
void RenderScene();        // Renders each frame.
void Shutdown();           // Releases all resources.


// The first is a global HDC that will be used in the RenderScene() function to swap buffers.
HDC g_HDC;


// xRotation and yRotation is used to rotate the entire scene.
float xRotation = 0.0f;
float yRotation = 0.0f;

// Texture object used to texture map a surface.
CTGALoader Texture;

// Arrays of all vertex and texture coordinate data.
float SquarePoints[] = {
   1.5f, -1.5f, 0.0f, -1.5f, -1.5f, 0.0f,
   -1.5f, 1.5f, 0.0f, 1.5f, 1.5f, 0.0f
};

float SquareTexCoords[] = {
   1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f
};

// Total triangle indices that make up our object.
int TotalTriangleIndices = 6;

// List of triangle indices.
unsigned int TriangleIndicesList[] = 
{
   // Triangle 1   // Triangle 2
   0, 1, 2,        2, 3, 0
};


// These are all the functions we will need for the AGP memory.
PFNGLBINDBUFFERARBPROC glBindBufferARB = NULL;
PFNGLDELETEBUFFERSARBPROC glDeleteBuffersARB = NULL;
PFNGLGENBUFFERSARBPROC glGenBuffersARB = NULL;
PFNGLBUFFERDATAARBPROC glBufferDataARB = NULL;
PFNGLGETBUFFERPARAMETERIVARBPROC glGetBufferParameterivARB = NULL;

// True if we allocated the buffer using the memory functions.
bool AGPAllocated = true;

// Just like with many other things in OpenGL, we need IDs to register
// (like with texture mapping).  This is used for the objects data.
unsigned int squarePointsBufferID = -1;
unsigned int squareTexCoordsBufferID = -1;
unsigned int squareElementID = -1;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                         // Rendering context.
   static HDC hDC;                           // Device context.
   int width, height;                        // The window width and height.
   static POINT oldMousePos;                 // Last mouse position.
   static POINT currentMousePos;             // Current mouse position.
   static bool  isMouseActive;               // Is the left mouse button down.

   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;
               }
            break;


         case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
				break;


			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
						xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;


         default:
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "AGP Memory (ARB_vertex_buffer_object) Allocating Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640,
                        480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }

   // Free all resources used by the demo.
   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

   glShadeModel(GL_SMOOTH);
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_TEXTURE_2D);

   // If the image failed to load we return false.
   if(!Texture.LoadTGA("UGP.tga")) return false;
   glGenTextures(1, &Texture.ID);
   
   glBindTexture(GL_TEXTURE_2D, Texture.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   int type = Texture.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, Texture.imageWidth,
                     Texture.imageHeight, type, GL_UNSIGNED_BYTE,
                     Texture.image);


   // With this we will load the list of extensions the hardware this demo is running on
   // supports.
   char *extension = (char*)glGetString(GL_EXTENSIONS);

   // Now that we have all of the extensions the hardware supports lets test if the
   // ARB_vertex_buffer_object extension is there.
   if(strstr(extension, "ARB_vertex_buffer_object") == NULL)
      {
         MessageBox(NULL, "ARB_vertex_buffer_object extension not supported!",
                    "Error...", MB_OK);

         return false;
      }

   // Next we load in all the OpenGL functions we need for AGP memory.
   glBindBufferARB = (PFNGLBINDBUFFERARBPROC)wglGetProcAddress("glBindBufferARB");
   glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC)wglGetProcAddress("glDeleteBuffersARB");
   glGenBuffersARB = (PFNGLGENBUFFERSARBPROC)wglGetProcAddress("glGenBuffersARB");
   glBufferDataARB = (PFNGLBUFFERDATAARBPROC)wglGetProcAddress("glBufferDataARB");
   glGetBufferParameterivARB = (PFNGLGETBUFFERPARAMETERIVARBPROC)wglGetProcAddress("glGetBufferParameterivARB");

   // Error checking...
   if(!glBindBufferARB || !glDeleteBuffersARB || !glGenBuffersARB || !glBufferDataARB ||
      !glGetBufferParameterivARB)
      {
         MessageBox(NULL,"Error getting one or more of the ARB vertex buffer object functions!",
                    "Error...", MB_OK);
         
         return false;
      }


   // This will determine the amount of space we need to draw everything in our scene.
   // Since this is just a square with 4 points we calculate it like so with
   // 3 components each...
   int AllocationSize = 4 * 3 * sizeof(float);
   int ArrayObjectSize = 0;

   // First we will generate the object ID then we bind it.  After that we will use
   // glBufferDataARB() to load the data of our object into the AGP memory.  Then we
   // will use glGetBufferParameterivARB to get the array size.  If the size <= 0 then
   // nothing was allocated on AGP memory (it failed).
   glGenBuffersARB(1, &squarePointsBufferID);
   glBindBufferARB(GL_ARRAY_BUFFER_ARB, squarePointsBufferID);
   glBufferDataARB(GL_ARRAY_BUFFER_ARB, AllocationSize, SquarePoints, GL_STATIC_DRAW_ARB);
   glGetBufferParameterivARB(GL_ARRAY_BUFFER_ARB, GL_BUFFER_SIZE_ARB, &ArrayObjectSize);


   // Error checking to make sure the object was loaded into AGP memory correctly...
   if(ArrayObjectSize <= 0)
      {
         MessageBox(NULL,"glBufferDataARB didn't allocate the object data into AGP memory!",
                    "Error...", MB_OK);
         AGPAllocated = false;
      }
   
   AllocationSize = 4 * 2 * sizeof(float);
   glGenBuffersARB(1, &squareTexCoordsBufferID);
   glBindBufferARB(GL_ARRAY_BUFFER_ARB, squareTexCoordsBufferID);
   glBufferDataARB(GL_ARRAY_BUFFER_ARB, AllocationSize, SquareTexCoords, GL_STATIC_DRAW_ARB);
   glGetBufferParameterivARB(GL_ARRAY_BUFFER_ARB, GL_BUFFER_SIZE_ARB, &ArrayObjectSize);


   // Error checking to make sure the object was loaded into AGP memory correctly...
   if(ArrayObjectSize <= 0)
      {
         MessageBox(NULL,"glBufferDataARB didn't allocate the object data into AGP memory!",
                    "Error...", MB_OK);
         AGPAllocated = false;
      }


   // Next we will load the triangle indices into AGP memory.  First we calculate the allocation
   // size of the indices like so...
   AllocationSize = sizeof(unsigned int) * TotalTriangleIndices;
   ArrayObjectSize = 0;

   // Now we will generate the element list ID that will hold the indicies then bind it.
   // After that we use glBufferDataARB() to load the triangle indices into memory.
   glGenBuffersARB(1, &squareElementID);
   glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, squareElementID);
   glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, AllocationSize, TriangleIndicesList, GL_STATIC_DRAW_ARB);
   glGetBufferParameterivARB(GL_ELEMENT_ARRAY_BUFFER_ARB, GL_BUFFER_SIZE_ARB, &ArrayObjectSize);


   // Error checking...
   if(ArrayObjectSize <= 0)
      {
         MessageBox(NULL,"glBufferDataARB didn't allocate indices into AGP memory!",
                    "Error...", MB_OK);
         AGPAllocated = false;
      }

   // Since we are done loading lets unbind everything.
   glBindBufferARB(GL_ARRAY_BUFFER_ARB, NULL);
   glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, NULL);

   // If all went well we return true.
   return true;
}


void RenderScene()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);    // Clears the screen.
   glLoadIdentity();                   // Reset modelview matrix for new frame.

   // Translate and rotate the object.
   glTranslatef(0.0f, 0.0f, -6.0f);
   glRotatef(-yRotation, 1.0f, 0.0f, 0.0f);
	glRotatef(-xRotation, 0.0f, 1.0f, 0.0f);
   
   // Bind the texture.
   glBindTexture(GL_TEXTURE_2D, Texture.ID);

   // Draw the object using the vertex arrays.
   if(AGPAllocated)
      {
         // Next for the vertex data.  First we bind the buffer that has
         // the data in it.  Then we call glVertexPointer() which will automatically
         // have access to the data.
         glBindBufferARB(GL_ARRAY_BUFFER_ARB, squarePointsBufferID);
         glVertexPointer(3, GL_FLOAT, 0, 0);
         glEnableClientState(GL_VERTEX_ARRAY);
         
         // Set up the texture coordinate pointer just like the vertex data.
         glBindBufferARB(GL_ARRAY_BUFFER_ARB, squareTexCoordsBufferID);
         glTexCoordPointer(2, GL_FLOAT, 0, 0);
         glEnableClientState(GL_TEXTURE_COORD_ARRAY);
         
         // Now we draw the object using an offset of 0 (we are starting from the front).
         glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, squareElementID);
         glDrawElements(GL_TRIANGLES, TotalTriangleIndices, GL_UNSIGNED_INT, 0);

         // Disable all client states we created.
         glDisableClientState(GL_VERTEX_ARRAY);
         glDisableClientState(GL_TEXTURE_COORD_ARRAY);

         // Unbind both arrays now that we are done.
         glBindBufferARB(GL_ARRAY_BUFFER_ARB, NULL);
         glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, NULL);
      }
   else
      {
         // We will use this if we couldn't allocate any AGP memory so those that did not support
         // the feature can still run the application.  This will only happen if there was an
         // error loading the data into AGP but AGP memory is supported (keep that in mind).
         glEnableClientState(GL_VERTEX_ARRAY);
         glVertexPointer(3, GL_FLOAT, 0, (void**)&SquarePoints[0]);
         glEnableClientState(GL_TEXTURE_COORD_ARRAY);
         glTexCoordPointer(2, GL_FLOAT, 0, (void**)&SquareTexCoords[0]);
         
         glDrawElements(GL_TRIANGLES, TotalTriangleIndices, GL_UNSIGNED_INT, TriangleIndicesList);
      }

   SwapBuffers(g_HDC);
}


void Shutdown()
{
   // This is done in the destructor but we are specifing to free the image ourselves anyway.
   Texture.FreeImage();
   glDeleteTextures(1, &Texture.ID);

   // Delete our IDs just like we do any other kind of ID in OpenGL.
   glDeleteBuffersARB(1, &squarePointsBufferID);
   glDeleteBuffersARB(1, &squareTexCoordsBufferID);
   glDeleteBuffersARB(1, &squareElementID);
}


// Recap:

// Neat.  Now we can have a hugh performance increase using AGP memory
// directly.  The only downfall to this is that you must manage the memory
// yourself and you are responsible for everything going good.  This is not that hard but can
// prove to be a challenge at the start.  So if you can write some good memory management code
// then using this will help your game engine get closer to being able to run at a professional
// level in no time.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com